
/*
 * $Id: Server.cc,v 1.44.2.10 1999/11/23 21:21:02 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <iostream.h>
#include <iomanip.h>

#include "Connection.h"
#include "ContentCfg.h"
#include "ContentMgr.h"
#include "ErrorMgr.h"
#include "LogComment.h"
#include "Rnd.h"
#include "RndPermut.h"
#include "Server.h"
#include "SrvConnMgr.h"
#include "SrvOpts.h"
#include "StatPhaseMgr.h"
#include "TblDistr.h"

#include "ServerSym.h"
#include "ContentSym.h"
#include "SocketSym.h"

#include "polyBcastChannels.h"
#include "polyLogTags.h"
#include "polyErrors.h"
#include "rndDistrStat.h"
#include "gadgets.h"
#include "globals.h"

ObjFarm<SrvXact> Server::TheXacts;

Server::Server(): theDirAccSel(0) {
	theConnMgr = new SrvConnMgr(this); // for now

	theChannels.append(TheLogCfgChannel);
	theChannels.append(TheLogStateChannel);
	startListen();
}

Server::~Server() {
	delete theConnMgr;
	deaf();
}

void Server::configure(const ServerSym *cfg, const NetAddr &aHost, bool bind2iface) {
	Assert(theConnMgr);

	SockOpt opt;
	Agent::configure(cfg, aHost, opt);

	configureContents(cfg);

	theHistory.window(TheSrvOpts.theHistoryWin);

	// setup listen socket

	Must(theSock.create());
	Must(theSock.blocking(false));
	Should(theSock.reuseAddr(true));
	Must(theSock.configure(opt));

	if (bind2iface)
		Must(theSock.bind(theHost));
	else
		Must(theSock.bind(NetAddr(0, theHost.port())));

	Must(theSock.listen());

	int acceptLmt = -1;
	cfg->acceptLmt(acceptLmt);
	theConnMgr->configure(opt, cfg->pconnUseLmt());
	theConnMgr->acceptLmt(acceptLmt);
	theConnMgr->idleTimeout(cfg->idlePconnTimeout());

	TheXacts.limit(1024); // magic, no good way to estimate
}

void Server::start() {
	Assert(theSock);
	Assert(theConnMgr);
	Agent::start();
	theReserv = TheFileScanner->setFD(theSock.fd(), dirRead, this);
}

void Server::noteXactDone(SrvXact *x) {
	Assert(x);
	theConnMgr->put(x->conn());
	TheXacts.put(x);
}

void Server::noteReadReady(int) {
	if (!theConnMgr->accept(theSock)) {	// fatal error
		deaf();
		FatalError(errServerGone);
	}
}

void Server::noteConnReady(Connection *conn) {
	startXact(conn); // that's it?!
}

void Server::noteLogEvent(BcastChannel *ch, OLog &log) {
	if (ch == TheLogCfgChannel) {
		log << bege(lgSrvCfg);
		// XXX: implement (SrvCfgRec)
		log << ende;
	} else
	if (ch == TheLogStateChannel) {
		log << bege(lgSrvState);
		log << theSeqId
			<< theHistory;
		log << ende;
	}
}

void Server::deaf() {
	if (theSock) {
		Comment(2) << "server " << theHost << " is closing listen socket " << theSock.fd() << endc;
		theSock.close();
	}
	if (theReserv)
		TheFileScanner->clearRes(theReserv);
}

void Server::startXact(Connection *conn) {
	SrvXact *x = TheXacts.get();
	x->exec(this, conn,
		theThinkDistr ? Time::Secd(theThinkDistr->trial()) : Time());
}

ContentCfg *Server::selectContent(const OidRec &or) {
	if (or.tid > 0)
		return TheContentMgr.get(or.tid);
	Assert(or.tid < 0);

	theDirAccSel->rndGen()->seed(RndPermut(or.oid, seqId()));
	const int tid = (int)theDirAccSel->trial();
	Assert(tid >= 0 && tid < theDirectAccess.count());
	return theDirectAccess[tid];
}

static
void mySyms2cfgs(Array<ContentSym*> &syms, Array<ContentCfg*> &cfgs) {
	for (int i = 0; i < syms.count(); ++i) {
		ContentCfg *cc = TheContentMgr.get(syms[i]);
		Check(cc->theSize);
		cfgs.append(cc);
	}
}

void Server::configureContents(const ServerSym *cfg) {
	Array<ContentCfg*> ccfgs;

	// get "global" content information
	Array<ContentSym*> csyms;
	Array<double> cprobs;
	if (!cfg->contents(csyms, cprobs) || !csyms.count()) {
		cerr << cfg->loc() << "no content specs for server " << cfg->kind() << endl;
		exit(-2);
	}
	Assert(csyms.count() == cprobs.count());
	mySyms2cfgs(csyms, ccfgs);

	// get direct access info
	Array<ContentSym*> dasyms;
	if (!cfg->directAccess(dasyms) || !dasyms.count()) {
		cerr << cfg->loc() << "no direct_access specs for server " << cfg->kind() << endl;
		exit(-2);
	}
	mySyms2cfgs(dasyms, theDirectAccess);

	// check that all direct access cfgs are unique
	// and build cprobsViaDa map
	Array<double> cprobsViaDa;
	cprobsViaDa.stretch(theDirectAccess.count());
	for (int d = 0; d < theDirectAccess.count(); ++d) {
		ContentCfg *da = theDirectAccess[d];
		int count = 0;
		for (int c = 0; c < ccfgs.count(); ++c) {
			if (ccfgs[c]->id() == da->id()) {
				cprobsViaDa.append(cprobs[c]);
				count++;
			}
		}
		if (count != 1) {
			cerr << dasyms[d]->loc() 
				<< "direct_access spec below counted " << count
				<< " times in the global contents specs (must be 1):"
				<< endl << *dasyms[d] << endl;
			exit(-2);
		}
	}

	// compute direct access probabilities using known content distribution
	double daProb = 0;
	for (int d = 0; d < theDirectAccess.count(); ++d)
		daProb += cprobsViaDa[d];
	Array<double> daprobs(theDirectAccess.count());
	for (int d = 0; d < theDirectAccess.count(); ++d)
		daprobs.append(cprobsViaDa[d]/daProb); // d's contribution

	// make selector for direct access content
	theDirAccSel = TblDistr::FromDistrTable("da-sel", daprobs);
	Assert(theDirAccSel);

	// calculated expected global content distribution
	// and average cachability of pages
	Array<double> expCProbs(ccfgs.count());
	double expCProbsSum = 0;
	for (int c = 0; c < ccfgs.count(); ++c) {
		const ContentCfg *cc = ccfgs[c];

		// compute average number of oids of type c generated per da
		double contPerDa = 0;
		for (int d = 0; d < theDirectAccess.count(); ++d) {
			ContentCfg *da = theDirectAccess[d];
			//cerr << here << "da[" << d << "]*" << daprobs[d] << '/' << daProb << ": " << da->compContPerCall(cc) << endl;
			contPerDa += da->compContPerCall(cc) * daprobs[d];
		}
		
		// check that content type is accessible
		if (contPerDa <= 0) {
			cerr << csyms[c]->loc()
				<< ": the content description below is not accessible "
				<< "(is not in direct_access closure):"
				<< endl << *csyms[c] << endl;
			exit(-2);
		}

		expCProbs[c] = contPerDa;
		expCProbsSum += contPerDa;
	}

	// XXX: this should be moved into Server::reportCfg() or similar
	static int repCount = 0;
	if (!repCount++) {

		// finilize calculation of expCProbs
		// and compute average cachability of pages
		double chbSum = 0;
		double sizeSum = 0;
		for (int c = 0; c < ccfgs.count(); ++c) {
			expCProbs[c] /= expCProbsSum;
			chbSum += expCProbs[c] * ccfgs[c]->theChbRatio;
			sizeSum += expCProbs[c] * RndDistrStat(ccfgs[c]->theSize).mean();
		}

		// report planned and expected global content distributions
		Comment(7) << "Content distribution on server " << theKind << ':' << endl;
		Comment << setw(15) << "content"
			<< ' ' << setw(15) << "planned%" 
			<< ' ' << setw(15) << "likely%" 
			<< ' ' << setw(15) << "error%" 
			<< endl;
		for (int c = 0; c < ccfgs.count(); ++c) {
			const ContentCfg *cc = ccfgs[c];
			Comment << setw(15) << (const char*)cc->kind()
				<< ' ' << setw(15) << 100*cprobs[c]
				<< ' ' << setw(15) << 100*expCProbs[c]
				<< ' ' << setw(15) << 100*Ratio(expCProbs[c]-cprobs[c], cprobs[c])
				<< endl;
		}
		Comment << "expected average cachability: " << 100*chbSum << '%' << endl;
		Comment << "expected average object size: " << sizeSum << "Bytes" << endl;
		Comment << endc;
	}
}


#ifdef POSSIBLE_FUTURE_CODE
	// for each [embedded] content type:
	//  Compute newContPerDa, the average number of [embedded]
	//   content oids generated during a DA, and
	//  Compute newContProb, probability of generating a new [embedded] 
	//   content oid when generating that content type oid
	// note: we do not skip da's, but for them newContProb[i] == 1.0
	for (int c = 0; c < ccfgs.count(); ++c) {
		ContentCfg *cc = ccfgs[c];

		// compute average number of oids of type c generated per da
		double contPerDa = 0;
		for (int d = 0; d < theDirectAccess.count(); ++d) {
			ContentCfg *da = theDirectAccess[d];
			//cerr << here << "da[" << d << "]*" << daprobs[d] << '/' << daProb << ": " << da->compContPerCall(cc) << endl;
			contPerDa += da->compContPerCall(cc) * daprobs[d];
		}
		
		// check that content type is accessible
		if (contPerDa <= 0) {
			cerr << csyms[c]->loc()
				<< ": the content description below is not accessible "
				<< "(is not in direct_access closure):"
				<< endl << *csyms[c] << endl;
			exit(-2);
		}

		// see above for the definitions
		const double newContPerDa = cprobs[c] / daProb;
		const double newContProb = newContPerDa / contPerDa;

		// cerr << here << "cc[" << cc->id() << "]*" << cprobs[c] << ": " << newContProb << " * " << contPerDa << " = " << newContPerDa << endl;

		if (newContProb > 1) {
			cerr << csyms[c]->loc()
				<< ": the content description below cannot be produced "
				<< "in large enough quantaties to achieve " << 100*cprobs[c]
				<< "% global contribution: "
				<< endl << *csyms[c] << endl;
			exit(-2);
		}

		// store in da's (like newContProb) instead!
		// otherwise servers cannot have different content distributions...
		// when doen, make "cc" pointer constant
		cc->newPerOid(newContPerDa);

		// notify containers (and others) about newCcProb
		for (int d = 0; d < theDirectAccess.count(); ++d) {
			ContentCfg *da = theDirectAccess[d];
			da->noteNewContProb(cc, newContProb); 
		}
	}	
#endif
