
/*
 * $Id: SrvXact.cc,v 1.31.2.9 2000/01/12 00:05:36 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <limits.h>
#include <strstream.h>

#include "Clock.h"
#include "Rnd.h"
#include "RndPermut.h"
#include "Connection.h"
#include "ContentCfg.h"
#include "ContBuildIter.h"
#include "EmbedContMdl.h"
#include "HttpDate.h"
#include "ErrorMgr.h"
#include "SharedOpts.h"
#include "Server.h"
#include "SrvXact.h"

#include "polyBcastChannels.h"
#include "polyErrors.h"
#include "gadgets.h"
#include "globals.h"


/* HTTP header strings (to avoid strlen() calls when creating headers) */

// status lines
static const String slHttp_10_200 = "HTTP/1.0 200 OK\r\n";       
static const String slHttp_10_304 = "HTTP/1.0 304 Not Modified\r\n";

// full header fields
static const String hfCcCachable  = "Cache-Control: public\r\n";
static const String hfCcUncachable= "Cache-Control: private,no-cache\r\n"
                                    "Pragma: no-cache\r\n";
static const String hfConnAlive   = "Connection: keep-alive\r\n";
static const String hfConnClose   = "Connection: close\r\n";

// header field prefixes
static const String hfpContLength = "Content-Length: ";
static const String hfpContType   = "Content-Type: ";
static const String hfpDate       = "Date: ";
static const String hfpExpires    = "Expires: ";
static const String hfpLmt        = "Last-Modified: ";
static const String hfpXSrvOids   = "X-Srv-Oids: ";
static const String hfpXSrvWid    = "X-Srv-Wid: ";
static const String hfpXXact      = "X-Xact: ";


void SrvXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theContentCfg = 0;
	theContBuildIter = 0;
	theReqHdr.reset();
	theTimes.reset();
	theRepHdrSize = -1;
	theProducedSize = 0;
	theSeed = -1;
}

Agent *SrvXact::owner() {
	return theOwner;
}

void SrvXact::exec(Server *anOwner, Connection *aConn, Time delay) {
	Assert(anOwner);
	theOwner = anOwner;
	theConn = aConn;
	if (delay > 0)
		sleepFor(delay);
	else
		doStart();
}

void SrvXact::doStart() {
	start(theConn);

	theState = stHdrWaiting;

	Assert(!theConn->theRd.theReserv); // cleared in conn mgr
	// note: we assume that conn mgr read() on first noteReadReady()
	noteDataReady(); // push state machine forward
}

void SrvXact::finish(int errNo) {
	if (theContBuildIter)
		theEmbedCount = theContBuildIter->embedCount();

	stop(errNo);
	if (theContBuildIter)
		theContentCfg->theEmbedCont->putIterator(theContBuildIter);
	theOwner->noteXactDone(this);
}

void SrvXact::wakeUp(const Alarm &a) {
	AlarmUser::wakeUp(a);
	doStart();
}

void SrvXact::noteReadReady(int) {
	Assert(theState == stHdrWaiting);

	theConn->read();

	if (theConn->bad())
		finish(-1);
	else
		noteDataReady();
}

void SrvXact::noteDataReady() {
	if (theReqHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {
		if (theConn->theRd.theReserv)
			theConn->theRd.stop(this);
		gotHeader();
		return;
	}

	if (theConn->theRdBuf.full()) { // too big header
		finish(errHugeHdr);
		return;
	}

	if (theConn->atEof()) {         // premature end of headers
		finish(errPrematureEoh);
		return;
	}

	if (!theConn->theRd.theReserv)  // need to read more
		theConn->theRd.start(this);
}

void SrvXact::gotHeader() {

	// dump request header
	if (TheOpts.theDumpFlags(dumpReq, dumpHdr))
		printMsg(theConn->theRdBuf, theReqHdr.theHdrSize);

	// note: clt and srv code assumes we consume only after parse!
	theConn->theRdBuf.consumed(theReqHdr.theHdrSize);

	if (theReqHdr.foreign()) {
		theOidRec.foreign = true;
		finish(errForeignReq); // foreign request
		return;
	}

	theOidRec = theReqHdr.theOidRec;

	// update history for this server if a direct public request
	if (theOidRec.direct() && theOidRec.wid == theOwner->id())
		theOwner->history().noteOid(theOidRec.oid);

	// reclaim failed oid if any
	if (theReqHdr.theFailedOid > 0)
		theOwner->history().noteOid(theReqHdr.theFailedOid);

	// note if the client will close the connection
	theConn->lastUse(!theReqHdr.keepAlive());

	// select content specs for this reply
	theContentCfg = theOwner->selectContent(theOidRec);

	theRepHdrSize = -1; // so that we know that we need to make a reply

	if (!theConn->theWr.theReserv)
		theConn->theWr.start(this);
	theState = stSpaceWaiting;
}

void SrvXact::noteWriteReady(int) {
	Assert(theState == stSpaceWaiting);
	WrBuf &buf = theConn->theWrBuf;

	if (theRepHdrSize < 0)
		makeRep(buf);

	Assert(theProducedSize > 0 || buf.contSize() > 0);
	
	// keep the write buffer full
	buf.pack();
	if (theContBuildIter)
		theContBuildIter->slice();
	else
		buf.appendRnd(Min(buf.spaceSize(),
			(Size)(theRepSize - theProducedSize - buf.contSize())));

	//cerr << here << "noteWriteReady: " << theProducedSize << " < " << theRepSize << endl;

	const Size sz = theConn->write();
	if (theConn->bad()) {
		finish(-1);
		return;
	}

	Assert(sz >= 0);
	theProducedSize += sz;
	if (theProducedSize >= theRepSize) { // done
		Assert(theProducedSize == theRepSize);
		finish(0);
		return;
	}

	if (!theConn->theWr.theReserv) // need write more
		theConn->theWr.start(this);
}

// decide what kind of reply to build
void SrvXact::makeRep(WrBuf &buf) {
	Assert(theContentCfg);

	// timestamps and cachability are used for both 304 and 200 replies
	theSeed = RndPermut(theOwner->seqId(), theOidRec.oid);
	theContentCfg->theObjLifeCycle.calcTimes(theSeed, theTimes);
	theOidRec.cachable = calcCachability();

	const char *repStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	if (theReqHdr.theIMS >= 0 && theTimes.lmt() >= 0 && 
	    theTimes.lmt() <= theReqHdr.theIMS)
		makeRep304(os);
	else
		makeRep200(os);

	buf.appended(os.tellp());

	Assert(theRepHdrSize > 0 && theRepSize >= theRepHdrSize);

	// dump reply header
	if (TheOpts.theDumpFlags(dumpRep, dumpHdr))
		printMsg(repStart, theRepHdrSize);
}

// build headers for a happy "200 OK" reply
void SrvXact::makeRep200(ostream &os) {
	os << slHttp_10_200;

	/*
	 * it is "good practice" to send general-header fields first,
	 * followed by request-header or response-header fields, and
	 * ending with the entity-header fields.
	 */

	/* general-header fields */

	os << (theOidRec.cachable ? hfCcCachable : hfCcUncachable);

	putStdFields(os);

	/* other entity-header fields */

	if (theTimes.showLmt())
		HttpDatePrint(os << hfpLmt, theTimes.lmt()) << crlf;

	const Size clen = calcRepSize();
	os << hfpContLength << (int)clen << crlf;

	if (theContentCfg->theMimeType)
		os << hfpContType << theContentCfg->theMimeType << crlf;

	/* extention-header fields (they are entity-header fields as well) */

	if (theReqHdr.theXactId)
		os << hfpXXact << theReqHdr.theXactId.genMutant() << crlf;

	// send URL generation data
	os << hfpXSrvWid << theOwner->id() << crlf;
	os << hfpXSrvOids << theOwner->history().secOid();
	if (theReqHdr.theOidShortage > 0) {
		int count = 0;
		const int newOid = theOwner->history().newOids(theReqHdr.theOidShortage, count);
		os << ':' << newOid << '.' << count;
	}
	os << crlf;

	os << crlf; // end-of-headers

	theRepHdrSize = os.tellp();
	theRepSize = theRepHdrSize + clen;
	if (theContentCfg->theEmbedCont) {
		theContBuildIter = theContentCfg->theEmbedCont->getIterator(clen, &theConn->theWrBuf);
		if (theContBuildIter)
			theContBuildIter->oid(theOidRec.oid, theSeed);
	}
}

void SrvXact::makeRep304(ostream &os) {
	os << slHttp_10_304;
	putStdFields(os);
	os << crlf; // end-of-headers

	theRepHdrSize = theRepSize = os.tellp();
}

// put fields acceptible for both 304 and 200 replies
void SrvXact::putStdFields(ostream &os) const {
	// general-header fields

	HttpDatePrint(os << hfpDate) << crlf;

	// indicate persistency (HTTP/1.0 defaults to non-persistent)
	if (theConn->reusable())
		os << hfConnAlive;

	// response-header fields (none)
	// entity-header fields

	if (theTimes.knownExp())
		HttpDatePrint(os << hfpExpires, theTimes.exp()) << crlf;
}

Size SrvXact::calcRepSize() const {
	// seed random generator and generate random size
	RndDistr &repSzD = *theContentCfg->theSize;
	repSzD.rndGen()->seed(theSeed);
	const double dh = repSzD();
	// prevent int overflows and leave room for the HTTP header
	const Size sz = (int)Min(ceil(dh), (double)INT_MAX - 100*1024);
	Assert(sz >= 0);
	return sz;
}

bool SrvXact::calcCachability() const {
	RndGen rng(theSeed);
	return rng.event(theContentCfg->theChbRatio);
}
